package de.uniulm.inkombio.pareto;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.filechooser.FileFilter;

@SuppressWarnings("serial")
final class MainWindow extends JFrame {
	
	/**
	 * Keeps track of the overall number of tabs.
	 */
	private int runcount = 0;
	
	private JComboBox<RNASequence> rnaSelection1;
	private JComboBox<RNASequence> rnaSelection2;
	private JCheckBox affineCheckBox;
	private JTabbedPane pane;
	private JButton loadButton;
	private JButton startButton;

	private JLabel matchingScoreLabel;
	private JLabel gapCostLabel;
	private JTextField gapCost;
	private JLabel gapStartCostLabel;
	private JTextField gapStartCost;

	/**
	 * Constructs the window and its elements.
	 */
	public MainWindow() {
		setDefaultCloseOperation(EXIT_ON_CLOSE);
		setSize(1280,800);
		setLocationRelativeTo(null);
		setTitle("Pareto-optimal RNA sequence-structure alignments by multi-objective dynamic programming");
		
		JPanel panelSequences = new JPanel();
		panelSequences.setLayout(new BorderLayout(10,10));
		panelSequences.setBorder(BorderFactory.createTitledBorder("Sequences"));
		JPanel panelInput = new JPanel();
		panelInput.setLayout(new GridLayout(3, 1, 2, 2));

		// Add first combo box
		JPanel p1 = new JPanel();
		p1.setLayout(new BorderLayout(10,10));
		p1.add(new JLabel("RNA 1:"), BorderLayout.WEST);
		rnaSelection1 = new JComboBox<RNASequence>();
		rnaSelection1.setToolTipText("Use this combo box to select the first RNA sequence from the database.");
		p1.add(rnaSelection1, BorderLayout.CENTER);
		panelInput.add(p1);
		
		// Add second combo box
		JPanel p2 = new JPanel();
		p2.setLayout(new BorderLayout(10,10));
		p2.add(new JLabel("RNA 2:"), BorderLayout.WEST);
		rnaSelection2 = new JComboBox<RNASequence>();
		rnaSelection2.setToolTipText("Use this combo box to select the second RNA sequence from the database.");
		p2.add(rnaSelection2, BorderLayout.CENTER);
		panelInput.add(p2);

		// Add configuration panel
		JPanel configurationAndInfoPanel = new JPanel();
		configurationAndInfoPanel.setLayout(new BorderLayout());
		configurationAndInfoPanel.add(new JLabel("Configuration:"), BorderLayout.WEST);
		
		JPanel configurationPanel = new JPanel();
		configurationPanel.setToolTipText("Set some parameters for the sequence alignment algorithm. The matching score is fixed and is 0.05 for matching bases.");
		configurationPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
		matchingScoreLabel = new JLabel("matching score: 0.05 (fixed)");
		matchingScoreLabel.setToolTipText(configurationPanel.getToolTipText());
		configurationPanel.add(matchingScoreLabel);
		configurationPanel.add(Box.createHorizontalStrut(10));
		affineCheckBox = new JCheckBox("use affine gap penalties");
		affineCheckBox.setToolTipText(configurationPanel.getToolTipText());
		affineCheckBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				boolean enabled = affineCheckBox.isSelected();
				gapStartCostLabel.setEnabled(enabled);
				gapStartCost.setEnabled(enabled);
			}
		});
		configurationPanel.add(affineCheckBox);
		configurationPanel.add(Box.createHorizontalStrut(10));
		gapCostLabel = new JLabel("Gap cost:");
		gapCostLabel.setToolTipText(configurationPanel.getToolTipText());
		configurationPanel.add(gapCostLabel);
		gapCost = new JTextField("-3.0");
		gapCost.setToolTipText(configurationPanel.getToolTipText());
		configurationPanel.add(gapCost);
		configurationPanel.add(Box.createHorizontalStrut(10));
		gapStartCostLabel = new JLabel("Gap opening cost:");
		gapStartCostLabel.setToolTipText(configurationPanel.getToolTipText());
		configurationPanel.add(gapStartCostLabel);
		gapStartCost = new JTextField("-3.0");
		gapStartCost.setToolTipText(configurationPanel.getToolTipText());
		configurationPanel.add(gapStartCost);
		configurationAndInfoPanel.add(configurationPanel, BorderLayout.CENTER);

		
		// Add version panel
		JPanel versionPanel = new JPanel();
		versionPanel.setLayout(new BorderLayout(10,10));
		JLabel versionLabel = new JLabel("RNA-Pareto " + RNAPareto.getVersion());
		JButton versionButton = new JButton("?");
		versionButton.setMargin(new Insets(1, 4, 1, 4));
		versionButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String message = String.format(RNAPareto.copyrightMessage,
						RNAPareto.getVersion());
				JOptionPane.showMessageDialog(MainWindow.this, message,
						"RNA-Pareto", JOptionPane.INFORMATION_MESSAGE);
			}
		});
		versionPanel.add(versionLabel, BorderLayout.CENTER);
		versionPanel.add(versionButton, BorderLayout.EAST);
		configurationAndInfoPanel.add(versionPanel, BorderLayout.EAST);
		
		panelInput.add(configurationAndInfoPanel);
		panelSequences.add(panelInput, BorderLayout.CENTER);

		// Add load button
		loadButton = new JButton("Load FASTA");
		loadButton.setToolTipText("Load a file of sequences in FASTA format.");
		loadButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JFileChooser fc = new JFileChooser();
				fc.setFileFilter(new FileFilter() {
					@Override
					public String getDescription() {
						return "FASTA-Files (*.fasta, *.fa)";
					}

					@Override
					public boolean accept(File f) {
						return f.isDirectory()
								|| f.getName().endsWith(".fasta")
								|| f.getName().endsWith(".fa");
					}
				});
				fc.setCurrentDirectory(new File("."));
				if (fc.showOpenDialog(MainWindow.this)==JFileChooser.APPROVE_OPTION) {
					String file = fc.getSelectedFile().getAbsolutePath();
					try {
						setFile(file);
					} catch (FileNotFoundException e1) {
						JOptionPane.showMessageDialog(MainWindow.this, "File " + file + " not found!", "Error", JOptionPane.ERROR_MESSAGE);
					}
				}
			}
		});
		panelSequences.add(loadButton, BorderLayout.WEST);
		
		// Add start button
		startButton = new JButton("Start");
		startButton.setToolTipText("Compute the set of Pareto-optimal sequence-structure alignments of the two selected sequences.");
		startButton.setEnabled(false);
		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				RNASequence a = (RNASequence) rnaSelection1.getSelectedItem();
				RNASequence b = (RNASequence) rnaSelection2.getSelectedItem();
				boolean affine = affineCheckBox.isSelected();
				double gap_score;
				try {
					gap_score = Double.parseDouble(gapCost.getText());
				} catch (NumberFormatException ex) {
					JOptionPane.showMessageDialog(MainWindow.this,
							"The field \"gap cost\" is not a number!",
							"Number format", JOptionPane.ERROR_MESSAGE);
					return;
				}
				double gap_start;
				try {
					gap_start = Double.parseDouble(gapStartCost.getText());
				} catch (NumberFormatException ex) {
					JOptionPane.showMessageDialog(MainWindow.this,
							"The field \"gap opening cost\" is not a number!",
							"Number format", JOptionPane.ERROR_MESSAGE);
					return;
				}
				if (gap_start > 0 || gap_score > 0) {
					JOptionPane.showMessageDialog(MainWindow.this,
							"The gap costs must be non-positive numbers!",
							"RNAPareto", JOptionPane.ERROR_MESSAGE);
					return;
				}
				actionStart(a, b, affine, gap_score, gap_start);
			}
		});
		panelSequences.add(startButton, BorderLayout.EAST);
		add(panelSequences, BorderLayout.NORTH);
		
		// Add tabbed pane and keystroke functionality
		pane = new JTabbedPane();
		add(pane, BorderLayout.CENTER);
		KeyStrokeFunctionality.addKeyStrokes(this, pane);
		setConfigurationEnabled(false);
	}
	
	/**
	 * After the user chooses a FASTA file to open, this method is called to
	 * update the combo boxes.
	 * 
	 * @param filename
	 *            file to open
	 * @throws FileNotFoundException 
	 */
	public void setFile(String filename) throws FileNotFoundException {
		List<RNASequence> list = RNAPareto.readFastaFile(filename);

		rnaSelection1.removeAllItems();
		rnaSelection2.removeAllItems();
		for (RNASequence r : list) {
			rnaSelection1.addItem(r);
			rnaSelection2.addItem(r);
		}

		if (list.size() > 0) {
			setConfigurationEnabled(true);
		} else {
			setConfigurationEnabled(false);
		}
	}
	
	/**
	 * Sets the enabled status of components for choosing RNAs and the
	 * configuration items.
	 * 
	 * @param enabled
	 */
	private void setConfigurationEnabled(boolean enabled) {
		startButton.setEnabled(enabled);
		rnaSelection1.setEnabled(enabled);
		rnaSelection2.setEnabled(enabled);
		matchingScoreLabel.setEnabled(enabled);
		affineCheckBox.setEnabled(enabled);
		gapStartCostLabel.setEnabled(affineCheckBox.isSelected());
		gapStartCost.setEnabled(affineCheckBox.isSelected());
		gapCostLabel.setEnabled(enabled);
		gapCost.setEnabled(enabled);
		if (enabled) {
			rnaSelection1.setSelectedIndex(0);
			rnaSelection2.setSelectedIndex(0);
		}
	}

	/**
	 * This method opens a new tab and invokes the calculation of the Pareto-set
	 * for the two RNAs a and b.
	 * 
	 * @param a
	 *            First RNA sequence
	 * @param b
	 *            Second RNA sequence
	 * @param affine
	 *            chooses the algorithm: normal gap costs or affine
	 * @param gap_score
	 *            Gap extension costs
	 * @param gap_start
	 *            Gap opening costs (only relevant for affine gap costs)
	 */
	void actionStart(RNASequence a, RNASequence b, boolean affine, double gap_score, double gap_start) {
		++runcount;
		String title = "Run #" + runcount;
		String tip = "\"" + a.name + "\" vs. \"" + b.name + "\"";

		final String a_sequence = a.sequence;
		final String b_sequence = b.sequence;

		final ParetoPanel p = new ParetoPanel(a, b, affine, gap_score, gap_start);
		pane.insertTab(title, null, p, tip, pane.getTabCount());
		pane.setTabComponentAt(pane.getTabCount() - 1, new ButtonTabComponent(
				pane));
		pane.setSelectedIndex(pane.getTabCount() - 1);

		final boolean f_affine = affine;
		final double f_gap_score = gap_score;
		final double f_gap_start = gap_start;

		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				Solution[] solutions;
				solutions = RNAPareto.getParetoAlignments(a_sequence,
						b_sequence, f_affine, f_gap_score, f_gap_start);
				ArrayList<Solution> solutionsList = new ArrayList<Solution>();
				if (solutions != null) {
					for (Solution s : solutions)
						solutionsList.add(s);
				}
				p.initData(solutionsList);
			}
		});
	}
	
}