package de.uniulm.inkombio.pareto;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import fr.orsay.lri.varna.VARNAPanel;
import fr.orsay.lri.varna.models.VARNAConfig;
import fr.orsay.lri.varna.models.VARNAConfig.BP_STYLE;
import fr.orsay.lri.varna.models.rna.RNA;

/**
 * A ParetoPanel contains the "result" of one run: the two text areas, the two
 * VARNA drawings and the scatter plot.
 * 
 * @author schnattinger
 * 
 */
@SuppressWarnings("serial")
public class ParetoPanel extends JPanel {

	private JTextArea top;
	private JPanel center;
	private JTextArea bottom;

	private VARNAPanel varnaPanel1;
	private VARNAPanel varnaPanel2;
	private ArrayList<Solution> solutions;
	ScatterPlotPanel scatterplot;

	private String rnaTitle1;
	private String rnaTitle2;
	private VARNAConfig config;
	private HidingSliderComponent slidercomponent;
	
	/**
	 * Contains the indices of the solutions which lie on the convex hull
	 */
	Map<Integer, Integer> convexhull;

	public ParetoPanel(RNASequence a, RNASequence b, boolean affine,
			double gap_score, double gap_start) {
		setLayout(new BorderLayout());
		top = new JTextArea(4, 80);
		top.setToolTipText("This text area shows the configuration details of the current run.");
		top.setEditable(false);
		top.setFont(new Font("Courier", 0, 12));
		add(new JScrollPane(top), BorderLayout.NORTH);
		JPanel main = new JPanel(new GridLayout(1, 3));
		add(main, BorderLayout.CENTER);
		bottom = new JTextArea(7, 80);
		bottom.setToolTipText("<html>"
				+ "This text area shows the details of the currently selected solution:<ul>"
				+ "<li>The first line contains the sequence and structure score.</li>"
				+ "<li>In the center there is the sequence-structure alignment of this solution.</li>"
				+ "<li>In the last line, the recalculated minimum free energies of the two RNAs, folded into the common consensus structure, is shown.</li>"
				+ "</ul></html>");
		bottom.setEditable(false);
		bottom.setFont(new Font("Courier", 0, 12));
		add(new JScrollPane(bottom), BorderLayout.SOUTH);

		varnaPanel1 = new VARNAPanel();
		varnaPanel1
				.setToolTipText("This drawing shows the first RNA sequence (\""
						+ a.name
						+ "\") featuring the consensus secondary structure of the current solution.");
		varnaPanel2 = new VARNAPanel();
		varnaPanel2
				.setToolTipText("This drawing shows the second RNA sequence (\""
						+ b.name
						+ "\") featuring the consensus secondary structure of the current solution.");

		center = new JPanel(new BorderLayout());
		JLabel wait = new JLabel("Compute Pareto-Front...");
		center.add(wait);

		main.add(varnaPanel1);
		main.add(center);
		main.add(varnaPanel2);

		String text = "Comparing:\n";
		text += "RNA 1: " + a.toString() + "\nRNA 2: " + b.toString() + "\n";
		text += "Use affine gap penalties: " + affine;
		text += ", ";
		if (affine) {
			text += "gap cost: " + gap_score + ", ";
			text += "gap opening cost: " + gap_start;
		} else {
			text += "uniform gap cost: " + gap_score;
		}
		top.setText(text);

		rnaTitle1 = a.name;
		rnaTitle2 = b.name;

		config = new VARNAConfig();
		config._flatExteriorLoop = false;
		config._mainBPStyle = BP_STYLE.SIMPLE;
		
		convexhull = new HashMap<Integer, Integer>();
	}

	/**
	 * Initialize this panel with the data points.
	 * 
	 * @param solutions
	 */
	public void initData(ArrayList<Solution> solutions) {
		this.solutions = solutions;
		
		if (solutions.size()==0) {
			center.removeAll();
			JLabel error = new JLabel("No alignments found... Maybe some error happened?");
			center.add(error);
			return;
		}

		// Initialize the internal points array
		double[][] d = new double[solutions.size()][];
		for (int i=0; i<d.length; i++) {
			d[i] = solutions.get(i).values;
		}
		double[] mfe_color = new double[solutions.size()];
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		for (int i=0; i<mfe_color.length; i++) {
			mfe_color[i] = -solutions.get(i).mfe;
			if (mfe_color[i]<0) mfe_color[i]=0;
			if (mfe_color[i]>max) max=mfe_color[i];
			if (mfe_color[i]<min) min=mfe_color[i];
		}
		if (max>min) {
			for (int i=0; i<mfe_color.length; i++) {
				mfe_color[i] -= min;
				mfe_color[i] /= (max-min);
				mfe_color[i] *= .5;
				mfe_color[i] += .5;
			}
		} else {
			for (int i=0; i<mfe_color.length; i++) {
				mfe_color[i] = .5;
			}
		}

		// Initialize and add the 2D plot.
		scatterplot = new ScatterPlotPanel("sequence score", "structure score", d, mfe_color, convexhull);
		scatterplot.setToolTipText("<html>Scatter plot of all Pareto-optimal solutions.<br>Use the keyboard or mouse to select a solution.</html>");
		scatterplot.setDataPointSelectionListener(new DataPointSelectionListener() {
			public void dataPointSelected(int index) {
				Solution s = ParetoPanel.this.solutions.get(index);
				StringBuilder text = new StringBuilder();
				text.append("Solution #").append(index+1).append(" ");
				text.append("(sequence score: ");
				text.append(String.format(Locale.US, "%.6f", s.values[0]));
				text.append(", structure score: ");
				text.append(String.format(Locale.US, "%.6f", s.values[1]));
				text.append("):\n\n");
				
				text.append(s.sequence_a).append("\n");
				text.append(s.sequence_b).append("\n");
				text.append(s.structure).append("\n\n");
				
				text.append("Minimum Free Energy 1: ");
				text.append(String.format(Locale.US, "%.2f", s.mfe_a)).append(" kcal/mol");
				text.append(", Minimum Free Energy 2: ");
				text.append(String.format(Locale.US, "%.2f", s.mfe_b)).append(" kcal/mol");
				
				bottom.setText(text.toString());
				
				setRNA(varnaPanel1, rnaTitle1, s.sequence_a, s.structure);
				setRNA(varnaPanel2, rnaTitle2, s.sequence_b, s.structure);
				
				if (slidercomponent!=null && !convexhull.containsKey(index)) {
					slidercomponent.hideMe();
				}
				if (slidercomponent!=null && convexhull.containsKey(index)) {
					slidercomponent.setValue(convexhull.get(index));
				}
			}
		});
		scatterplot.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				super.mouseClicked(e);
				if (scatterplot.getNearestDataPointNode(e.getX(), e.getY(), 5) < 0) {
					slidercomponent.hideMe();
				}
			}
		});
		
		center.removeAll();
		center.add(scatterplot, BorderLayout.CENTER);
		slidercomponent = new HidingSliderComponent(solutions, scatterplot, this);
		center.add(slidercomponent, BorderLayout.SOUTH);
		center.validate();
		
		scatterplot.selectDataPointNode(0);
	}

	/**
	 * Draw the RNA (sequence,structure) into the VARNAPanel _vp
	 * 
	 * @param _vp
	 * @param title
	 * @param sequence
	 * @param structure
	 */
	protected void setRNA(VARNAPanel _vp, String title, String sequence,
			String structure) {
		int length = 0;
		for (int i = 0; i < sequence.length(); i++)
			if (sequence.charAt(i) != '-')
				++length;
		char[] seq = new char[length];
		char[] str = new char[length];
		int p = 0;
		for (int i = 0; i < sequence.length(); i++) {
			if (sequence.charAt(i) != '-') {
				seq[p] = sequence.charAt(i);
				str[p] = structure.charAt(i);
				++p;
			}
		}

		try {
			RNA r = new RNA(title);
			r.setRNA(new String(seq), new String(str));
			r.drawRNARadiate(config);

			// Why ever. Flip the RNA horizontally, else RNA
			// display incorrect. This is VARNA's fault.
			// ---------------------------------------
			for (int i = 0; i < r.getSize(); i++) {
				Point2D.Double a = r.getCoords(i);
				r.setCoord(i, a.x, -a.y);
			}
			// ---------------------------------------

			_vp.showRNA(r, config);
			_vp.repaint();
		} catch (Exception e1) {
			JOptionPane.showMessageDialog(_vp, e1.getMessage(), "Error",
					JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * selects the data point r points ahead
	 * 
	 * @param r
	 */
	public void selectDataPointNodeRelative(int r) {
		scatterplot.selectDataPointNodeRelative(r);
	}
}
