package de.uniulm.inkombio.pareto;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.UIManager;

/**
 * RNAPareto is the main class as well as the Java interface to the native JNI
 * implementation of RNA-Pareto.
 * 
 * @author schnattinger
 * 
 */
public class RNAPareto {
	
	/**
	 * This is the HTML copyright message which is displayed on push on the '?'
	 * button
	 */
	public static final String copyrightMessage = "<html><u><i>RNA-Pareto %s</i></u><br><br>"
			+ "This program is written by Thomas Schnattinger at Ulm University, 89069 Ulm, Germany. <br>"
			+ "It comes in binary format for Ubuntu Linux, Windows XP/7 and Mac, as well as source code.<br>"
			+ "It is free software under the terms of the GNU GPLv3 license.<br><br><br>"
			+ "Use of third party software:<ul>"
			+ "<li>The calculation of the base pair probabilities and the minimum free energies is done by<br>"
			+ "<i>ViennaRNA Package</i> (<a href=\"http://www.tbi.univie.ac.at/~ronny/RNA/\">http://www.tbi.univie.ac.at/~ronny/RNA/</a>)</li>"
			+ "<li>The drawing of the RNAs is done by<br>"
			+ "<i>VARNA: Visualization Applet for RNA</i> (<a href=\"http://varna.lri.fr/\">http://varna.lri.fr/</a>)</li>"
			+ "</ul><br>"
			+ "References:<ul>"
			+ "<li>Schnattinger, T., Schöning, U., and Kestler, H. A. (2013). <i>Structural RNA alignment by multi-objective optimization</i>.<br>Bioinformatics.</li>"
			+ "<li>Schnattinger, T., Schöning, U., and Kestler, H. (2012). <i>Pareto-optimal RNA sequence-structure alignments</i>.<br>In 9th International Workshop on Computational Systems Biology 2012 (WCSB 2012), pages 83–86, Ulm, Germany.</li>"
			+ "<li>Lorenz, R., Bernhart, S. H., zu Siederdissen, C. H., Tafer, H., Flamm, C., Stadler, P. F., and Hofacker, I. L. (2011).<br><i>ViennaRNA Package 2.0</i>. Algorithms for Molecular Biology, 6(26).</li>"
			+ "<li>Darty, K., Denise, A., and Ponty, Y. (2009). <i>VARNA: Interactive drawing and editing of the RNA secondary structure</i>.<br>Bioinformatics, 25(15), 1974–1975.</li>"
			+ "</ul><br>"
			+ "Homepage: <a href=\"http://sysbio.uni-ulm.de?Software\">http://sysbio.uni-ulm.de?Software</a><br><br>"
			+ "Contact: <a href=\"mailto:thomas.schnattinger@uni-ulm.de\">"
			+ "thomas.schnattinger@uni-ulm.de</a><br><br><hr />"
			+ "<center><table border=\"0\"><tr><td rowspan=\"5\" valign=\"top\">Control keys:</td>"
			+ "<td>Ctrl+Up</td><td>Select previous data point</td><td>&nbsp;</td><td>Ctrl+PgUp</td><td>Switch to previous tab</td></tr>"
			+ "<tr><td>Ctrl+Down</td><td>Select next data point</td><td>&nbsp;</td><td>Ctrl+PgDown</td><td>Switch to next tab</td></tr>"
			+ "<tr><td>Ctrl+Home</td><td>Select first data point</td><td>&nbsp;</td><td>Ctrl+W</td><td>Close current tab</td></tr>"
			+ "<tr><td>Ctrl+End</td><td>Select last data point</td><td>&nbsp;</td><td>Ctrl+Q</td><td>Quit program</td></tr>"
			+ "<tr><td>Mouseclick</td><td>Select data point</td><td>&nbsp;</td><td>Mousewheel</td><td>Switch between data points</td></tr>"
			+ "</table></center></html>";

	/**
	 * Tells if the shared library could be loaded successfully or not.
	 */
	public static final boolean libraryPresent;
	
	/**
	 * System specific library name (for error messages).
	 */
	public static final String libraryFileName;
	
	static {
		// try to load the library from different locations...
		boolean success = true;
		try {
			System.load(System.getProperty("user.dir")
					+ System.getProperty("file.separator")
					+ System.mapLibraryName("rna-pareto"));
		} catch (UnsatisfiedLinkError e) {
			try {
				System.loadLibrary("rna-pareto");
			} catch (UnsatisfiedLinkError f) {
				success = false;
			}
		}
		libraryPresent = success;
		libraryFileName = System.mapLibraryName("rna-pareto");
	}
	
	static {
		// try to load system look and feel, but continue if it does not work
		// out
		String systemLookAndFeel = UIManager.getSystemLookAndFeelClassName();
		try {
			UIManager.setLookAndFeel(systemLookAndFeel);
		} catch (Exception e) {
			System.err.println("Info: Java System look and feel ("
					+ systemLookAndFeel + ") could not be set.");
		}
	}

	/**
	 * Returns an array of pareto-optimal Solutions to the sequence-structure
	 * alignment problem. Implemented by an external JNI library.
	 * 
	 * @param sequence_a
	 *            The first RNA sequence
	 * @param sequence_b
	 *            The second RNA sequence
	 * @param affine
	 *            Flag for the alignment method
	 * @param gap_score
	 *            Gap extension costs
	 * @param gap_start
	 *            Gap opening costs (only relevant for affine alignment)
	 * 
	 * @return Array of pareto-optimal solutions.
	 */
	public static native Solution[] getParetoAlignments(String sequence_a,
			String sequence_b, boolean affine, double gap_score, double gap_start);
	
	/**
	 * Get the library version.
	 * 
	 * @return Version major.minor.revision
	 */
	public static native String getVersion();

	/**
	 * The main method of this tool.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		
		// If the shared library can not be loaded, stop here.
		if (!libraryPresent) {
			String message = "Error: The shared library '" + libraryFileName + "' could not be loaded!\n\n" + 
					"It has to be present in either the same directory as the jar file or in 'java.library.path'.\n" + 
					"Please contact thomas.schnattinger@uni-ulm.de for help!";
			JOptionPane.showMessageDialog(null, message, "RNAPareto : Error", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}
		
		MainWindow f = new MainWindow();
		f.setVisible(true);
		f.setExtendedState(f.getExtendedState()|JFrame.MAXIMIZED_BOTH);
	}

	/**
	 * This method reads a file in FASTA format and returns a List of
	 * RNASequence objects.
	 * 
	 * @param filename input file in FASTA format
	 * @return List of RNASequence objects
	 * @throws FileNotFoundException
	 */
	public static List<RNASequence> readFastaFile(String filename)
			throws FileNotFoundException {
		ArrayList<RNASequence> ret = new ArrayList<RNASequence>();
		BufferedReader br = new BufferedReader(new FileReader(filename));
		try {
			String line;
			String name = null;
			StringBuilder sequence = new StringBuilder();
			do {
				line = br.readLine();
				if (line == null || line.trim().startsWith(">")) {
					if (sequence.length() > 0) {
						ret.add(new RNASequence(name, sequence.toString()
								.replaceAll("[^ACGUTacgut]", "")));
					}
					if (line != null)
						name = line.trim().substring(1).trim();
					sequence.setLength(0);
				} else {
					sequence.append(line);
				}
			} while (line != null);
			br.close();
		} catch (IOException e) {
			e.printStackTrace(System.err);
		}
		return ret;
	}
}
