package de.uniulm.inkombio.pareto;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;

/**
 * This Panel paints a nice scatter plot of the given array of points. Own
 * implementation because Gnuplot/Javaplot/SVG was not portable enough.
 */
@SuppressWarnings("serial")
public class ScatterPlotPanel extends JPanel {

	private DataPointSelectionListener dataPointSelectionListener;
	
	/**
	 * Caption on the horizontal axis
	 */
	String xCaption;
	
	/**
	 * Caption on the vertical axis
	 */
	String yCaption;

	/**
	 * List of the solutions' coordinates in [0,1]x[0,1] 
	 */
	private List<Point2D.Double> points;
	
	/**
	 * Color of the solutions (for MFE)
	 */
	private List<Color> pointColors;

	/**
	 * Index of currently selected solution
	 */
	protected int currentlySelected = -1;

	/**
	 * The rectangle in which the data is painted
	 */
	private Rectangle2D.Double paintRange;

	/**
	 * List of the labels on the x-axis
	 */
	private List<String> xLabel;
	
	/**
	 * List of positions of the labels on the x-axis
	 */
	private List<Double> xLabelPosition;
	
	/**
	 * List of the labels on the y-axis
	 */
	private List<String> yLabel;

	/**
	 * List of positions of the labels on the y-axis
	 */
	private List<Double> yLabelPosition;
	
	/**
	 * Font for the labels (System default font)
	 */
	private final Font normalFont = getFont();
	
	/**
	 * Font for the axis captions (bold font)
	 */
	private final Font boldFont = normalFont.deriveFont(Font.BOLD);
	
	private final Font rotatedBoldFont = boldFont.deriveFont(AffineTransform.getRotateInstance(3*Math.PI/2));

	/**
	 * Maps the points of the convex hull to their corresponding weighting
	 */
	private Map<Integer, Integer> convexhull;

	private boolean displayConvexHull;

	/**
	 * Calculates an appropriate step size for dividing an interval of size m
	 * into a given number of steps.
	 * 
	 * @param m interval size
	 * @param steps number of steps
	 * @return appropriate step size
	 */
	private static double calculateAxisIntervals(double m, int steps) {
		double stepsize = m / steps;
		double log = Math.pow(10, Math.floor(Math.log10(stepsize)));
		stepsize = Math.round(stepsize/log)*log;
		return stepsize;
	}
	
	/**
	 * Constructor for this scatter plot panel.
	 * 
	 * @param d Array of coordinates of the solutions
	 * @param mfe_color Array of mfe values (between 0 and 1) of the solutions
	 */
	public ScatterPlotPanel(String xCaption, String yCaption, double[][] d, double[] mfe_color, Map<Integer, Integer> convexhull) {
		this.xCaption = xCaption;
		this.yCaption = yCaption;
		this.convexhull = convexhull;
		
		initData(d, mfe_color);
		
		MouseAdapter ml = new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				selectDataPointNode(getNearestDataPointNode(e.getX(), e.getY(), 5));
				super.mouseClicked(e);
			}
			@Override
			public void mouseWheelMoved(MouseWheelEvent e) {
				int r = -e.getWheelRotation();
				selectDataPointNode(currentlySelected+r);
			}
		};
		addMouseListener(ml);
		addMouseWheelListener(ml);
	}
	
	private void initData(double[][] d, double[] mfe_color) {
		double minX = Double.MAX_VALUE;
		double maxX = -Double.MAX_VALUE;
		double minY = Double.MAX_VALUE;
		double maxY = -Double.MAX_VALUE;
		
		for (int i = 0; i < d.length; i++) {
			if (d[i][0] < minX) minX = d[i][0];
			if (d[i][0] > maxX) maxX = d[i][0];
			if (d[i][1] < minY) minY = d[i][1];
			if (d[i][1] > maxY) maxY = d[i][1];
		}
		
		if (minX==maxX) {
			minX -= 5;
			maxX += 5;
		}
		
		if (minY==maxY) {
			minY -= 5;
			maxX += 5;
		}

		double xStep = calculateAxisIntervals(maxX-minX, 8);
		double yStep = calculateAxisIntervals(maxY-minY, 8);
		
		Rectangle2D.Double dataRange = new Rectangle2D.Double(minX, minY, maxX
				- minX, maxY - minY);

		xLabel = new ArrayList<String>();
		xLabelPosition = new ArrayList<Double>();
		for (double x = Math.ceil(minX/xStep)*xStep; x<=Math.floor(maxX/xStep)*xStep; x+=xStep) {
			xLabel.add(String.valueOf((float)x));
			xLabelPosition.add((x - dataRange.x) / dataRange.width);
		}
		
		yLabel = new ArrayList<String>();
		yLabelPosition = new ArrayList<Double>();
		for (double y = Math.ceil(minY/yStep)*yStep; y<=Math.floor(maxY/yStep)*yStep; y+=yStep) {
			yLabel.add(String.valueOf((float)y));
			yLabelPosition.add((y - dataRange.y) / dataRange.height);
		}
		
		points = new ArrayList<Point2D.Double>();
		pointColors = new ArrayList<Color>();
		for (int i = 0; i < d.length; i++) {
			points.add(new Point2D.Double((d[i][0] - dataRange.x)
					/ dataRange.width, (d[i][1] - dataRange.y)
					/ dataRange.height));
			pointColors.add(new Color((float) 1.0, (float) (1.0 - mfe_color[i]),
							(float) (1.0 - mfe_color[i])));
		}
	}

	public void paintComponent(Graphics g) {
		Graphics2D g2d = (Graphics2D)g;
		g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		
		int width = getWidth();
		int height = getHeight();
		
		paintRange = new Rectangle2D.Double(90, 20, width-110, height-90);
		
		// Draw white background
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, width, height);
		
		// Draw black border
		g.setColor(Color.BLACK);
		g.drawRect((int) (paintRange.x - 10), (int) (paintRange.y - 10),
				(int) (paintRange.width + 20), (int) (paintRange.height + 20));
		
		// Draw X axis
		g.setFont(normalFont);
		for (int i = 0; i < xLabel.size(); i++) {
			int x = (int) (paintRange.x + paintRange.width
					* xLabelPosition.get(i));
			g.drawLine(x, 7, x, 13);
			g.drawLine(x, (int) (paintRange.y + paintRange.height + 7), x,
					(int) (paintRange.y + paintRange.height + 13));
			String label = xLabel.get(i);
			g.drawString(label,
					x - SwingUtilities.computeStringWidth(g.getFontMetrics(), label) / 2,
					(int) (paintRange.y + paintRange.height + 35));
		}
		g.setFont(boldFont);
		g.drawString(xCaption,
				(int) (paintRange.x + (paintRange.width - SwingUtilities
						.computeStringWidth(g.getFontMetrics(), xCaption)) / 2),
				(int) (paintRange.y + paintRange.height + 60));
		
		// Draw Y axis
		g.setFont(normalFont);
		for (int i = 0; i < yLabel.size(); i++) {
			int y = (int) (paintRange.y + paintRange.height
					* (1 - yLabelPosition.get(i)));
			g.drawLine((int)paintRange.x-7, y, (int)paintRange.x-13, y);
			g.drawLine((int) (paintRange.x + paintRange.width + 7), y,
					(int) (paintRange.x + paintRange.width + 13), y);
			String label = yLabel.get(i);
			g.drawString(label,
					(int)paintRange.x - 20 - SwingUtilities.computeStringWidth(g.getFontMetrics(), label),
					y + (g.getFontMetrics().getAscent() / 2));
		}
		g.setFont(rotatedBoldFont);
		g.drawString(yCaption, (int) (paintRange.x - 60),
				(int) (paintRange.y + (paintRange.height + SwingUtilities
						.computeStringWidth(g.getFontMetrics(boldFont), yCaption)) / 2));
		
		g.setFont(normalFont);
		
		// Highlight convex hull
//		if (displayConvexHull) {
//			for (int i : convexhull.keySet()) {
//				Point2D.Double pt = points.get(i);
//				g.setColor(new Color(255, 180, 180));
//				g.fillOval((int) (-8 + paintRange.x + paintRange.width * pt.x),
//						(int) (-8 + paintRange.y + paintRange.height * (1 - pt.y)),
//						16, 16);
//				g.drawOval((int) (-8 + paintRange.x + paintRange.width * pt.x),
//						(int) (-8 + paintRange.y + paintRange.height * (1 - pt.y)),
//						16, 16);
//				
//			}
//		}
		
		// Draw data points
		for (int i=0; i<points.size(); i++) {
			Point2D.Double pt = points.get(i);
			if (i==currentlySelected) {
				g.setColor(new Color(0, 0, 128));
				g.fillOval((int) (-4 + paintRange.x + paintRange.width * pt.x),
						(int) (-4 + paintRange.y + paintRange.height * (1 - pt.y)),
						8, 8);
				g.drawOval((int) (-4 + paintRange.x + paintRange.width * pt.x),
						(int) (-4 + paintRange.y + paintRange.height * (1 - pt.y)),
						8, 8);
			} else {
				if (displayConvexHull && convexhull.containsKey(i)) {
					g.setColor(new Color(64,64,192));
				} else {
					g.setColor(pointColors.get(i));
				}
				g.drawOval((int) (-4 + paintRange.x + paintRange.width * pt.x),
						(int) (-4 + paintRange.y + paintRange.height * (1 - pt.y)),
						8, 8);
			}
		}
		g.setFont(normalFont);
		
	}
	
	/**
	 * Get the index of the nearest data point node from display coordinates (x,y).
	 * 
	 * @param x
	 * @param y
	 * @return
	 */
	public int getNearestDataPointNode(int x, int y, double maxDistance) {
		int nearest = -1;
		double dx = (x - paintRange.x) / paintRange.width;
		double dy = 1 - (y - paintRange.y) / paintRange.height;
		maxDistance /= Math.sqrt(paintRange.width*paintRange.height);
		for (int i=0; i<points.size(); i++) {
			Point2D.Double pt = points.get(i);
			double d = Math.sqrt((dx-pt.x)*(dx-pt.x)+(dy-pt.y)*(dy-pt.y));
			if (d<maxDistance) {
				maxDistance = d;
				nearest = i;
			}
		}
		return nearest;
	}
	
	/**
	 * Selects the 
	 * 
	 * @param newPoint
	 */
	public void selectDataPointNode(int newPoint) {
		if (newPoint>=0 && newPoint<points.size()) {
			currentlySelected = newPoint;
			repaint();
			if (dataPointSelectionListener!=null) {
				dataPointSelectionListener.dataPointSelected(newPoint);
			}
		}
	}
	
	/**
	 * selects the data point r points ahead
	 * 
	 * @param r
	 */
	public void selectDataPointNodeRelative(int r) {
		if (points.size()==0) return;
		int c = currentlySelected+r;
		if (c<0) c=0;
		else if (c>=points.size()) c=points.size()-1;
		selectDataPointNode(c);
	}

	/**
	 * Set a data point selection listener.
	 * 
	 * @param l
	 */
	public void setDataPointSelectionListener(DataPointSelectionListener l) {
		dataPointSelectionListener = l;
	}
	
	/**
	 * 
	 * @param display
	 */
	public void setDisplayConvexHull(boolean display) {
		displayConvexHull = display;
		repaint();
	}
}