package de.uniulm.inkombio.pareto;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

@SuppressWarnings("serial")
public class SliderComponent extends JPanel {

	private JSlider slider;
	private JLabel leftlabel;
	private JLabel rightlabel;
	private double value;
	
	private static final String tooltip = "Find the solutions on the Pareto-front which maximize the objective function for a given weight.";

	/**
	 * This array holds the indices of the solution that maximizes the objective
	 * function for a specific weighting value
	 */
	private int[] optimalSolution;
	
	/**
	 * Maps the indices of the points on the convex hull to their corresponding
	 * optimal weighting
	 */
	public Map<Integer, Integer> convexhull;

	private ScatterPlotPanel scatterplot;
	private boolean disableListening = false;

	/**
	 * Constructor for this slider component. Side effect: fills the parameter
	 * "convexhull" with the points on the convex hull (in the form
	 * "index:optimal weight").
	 * 
	 * @param title
	 * @param solutions
	 * @param scatterplot
	 * @param convexhull
	 */
	public SliderComponent(String title, ArrayList<Solution> solutions,
			ScatterPlotPanel scatterplot, Map<Integer, Integer> convexhull) {
		this.scatterplot = scatterplot;
		this.convexhull = convexhull;

		precalcluateOptimaForSlider(solutions);

		setBackground(Color.WHITE);
		setBorder(BorderFactory.createTitledBorder(title));
		setLayout(new BorderLayout());

		JPanel contentpane = new JPanel();
		contentpane.setBackground(Color.WHITE);
		contentpane.setBorder(new EmptyBorder(10, 10, 10, 10));
		contentpane.setLayout(new BorderLayout(10, 10));
		add(contentpane);

		leftlabel = new JLabel("0000");
		rightlabel = new JLabel("0000");
		contentpane.add(leftlabel, BorderLayout.WEST);
		contentpane.add(rightlabel, BorderLayout.EAST);

		slider = new JSlider(0, 1000, 0);
		slider.setBackground(Color.WHITE);
		slider.setPaintLabels(false);
		slider.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent e) {
				valueChanged();
			}
		});
		contentpane.add(slider, BorderLayout.CENTER);
		
		setToolTipText(tooltip);
		leftlabel.setToolTipText(tooltip);
		rightlabel.setToolTipText(tooltip);
		slider.setToolTipText(tooltip);
		
		valueChanged();
	}

	/**
	 * Precaluculate the optima arrays for the slider component.
	 * 
	 * @param solutions
	 */
	private void precalcluateOptimaForSlider(ArrayList<Solution> solutions) {
		optimalSolution = new int[1001];
		for (int i = 0; i <= 1000; i++) {
			double v = (1.0 - i / 1000.0);
			double maxresult = -Double.MAX_VALUE;
			int max_j = -1;
			for (int j = 0; j < solutions.size(); j++) {
				Solution s = solutions.get(j);
				double result = s.values[0] * v + s.values[1] * (1.0 - v);
				if (max_j < 0 || result > maxresult) {
					maxresult = result;
					max_j = j;
				}
			}
			optimalSolution[i] = max_j;
			convexhull.put(max_j, i);
		}
	}

	/**
	 * This method gets called when the value of the slider has changed.
	 */
	protected void valueChanged() {
		value = 1.0 * slider.getValue() / slider.getMaximum();
		leftlabel.setText("[sequence] "
				+ String.format(Locale.US, "%.2f", 1.0 - value));
		rightlabel.setText(String.format(Locale.US, "%.2f", value)
				+ " [structure]");
		if (!disableListening) {
			scatterplot.selectDataPointNode(optimalSolution[slider.getValue()]);
		}
	}

	/**
	 * Sets the value of the slider, but deactivates some functionality of the
	 * valueChanged() method before.
	 * 
	 * @param value
	 */
	public void setValue(int value) {
		disableListening  = true;
		slider.setValue(value);
		disableListening = false;
	}

	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);

	}
}
