
#include <iostream>
#include <iomanip>
#include <vector>
#include <stack>
#include <list>

#include "alignment.hpp"

// RNAeval, defined in rnafold-wrapper.c
extern "C" double RNAeval(const char* sequence, const char* structure);

match::match(const int type, const int n) : type(type), i(n), j(n), k(n), l(n) {}

match::match(const int type, const int i, const int k) : type(type), i(i), j(i), k(k), l(k) {}

match::match(const int type, const int i, const int j, const int k, const int l) : type(type), i(i), j(j), k(k), l(l) {}

match::match(const match &m) : type(m.type), i(m.i), j(m.j), k(m.k), l(m.l) {}

bool match::operator<(const match& m) const {
	if (l==m.l) {
		if (j==m.j) {
			return type>m.type;
		} else return j>m.j;
	} else return l>m.l;
}

alignment& alignment::operator= (const alignment& a) {
	if (this != &a) {
		alignment tmp(a);
		std::swap(*this, tmp);
	}
	return *this;
}

alignment::alignment(const std::vector<double> &objectives, const std::string &sequence_a, const std::string &sequence_b, const std::string &structure)
		: objectives(objectives), sequence_a(sequence_a), sequence_b(sequence_b), structure(structure) {}

void alignment::print_alignment() {
	for (std::vector<double>::const_iterator it=objectives.begin(); it!=objectives.end(); ++it) {
		if (it!=objectives.begin()) std::cout << " ";
		std::cout << std::setprecision(6) << std::fixed << *it;
	}
	std::cout << std::endl << sequence_a << std::endl << sequence_b << std::endl << structure << std::endl << std::endl;
}

double alignment::energy_of_structure(const std::string &sequence, const std::string &structure) {
	int i, j;
	char* seq = new char[sequence.length()+1];
	char* str = new char[structure.length()+1];
	for (i=0, j=0; i<sequence.length(); i++) {
		if (sequence[i]!='-') {
			seq[j] = sequence[i];
			str[j] = structure[i];
			j++;
		}
	}
	seq[j] = str[j] = '\0';
	double mfe = RNAeval(seq, str);
	delete[] seq;
	delete[] str;
	return mfe;
}

void print_matches(std::vector<match> &result) {
	for (std::vector<match>::iterator it=result.begin(); it!=result.end(); ++it) {
		std::cout << "match [" << it->i << " " << it->j << " " << it->k << " " << it->l << "] type=" << it->type << std::endl;
	}
	std::cout << std::endl;
}

void sort_matches(std::vector<match> &v) {
	// vector of matched
	std::vector<match> m;
	// vector of unmatched
	std::vector<match> u;
	for (std::vector<match>::iterator it=v.begin(); it!=v.end(); ++it) {
		if (it->type==1 || it->type==2) {
			u.push_back(*it);
		} else {
			m.push_back(*it);
		}
	}
	std::sort(m.begin(), m.end());
	std::list<match> l(m.begin(), m.end());
	for (std::vector<match>::iterator it1=u.begin(); it1!=u.end(); ++it1) {
		if (it1->type==1) {
//			std::cout << "insert (" << it1->j << ", -)" << std::endl;
			// unmatched j !!
			for (std::list<match>::iterator it2=l.begin(); it2!=l.end(); ++it2) {
				if (it2->j < it1->j) {
					l.insert(it2, *it1);
					goto nextelement;
				}
			}
			l.insert(l.end(), *it1);
		} else {
//			std::cout << "insert (-, " << it1->l << ")" << std::endl;
			// unmatched l !!
			for (std::list<match>::iterator it2=l.begin(); it2!=l.end(); ++it2) {
				if (it2->l < it1->l) {
					l.insert(it2, *it1);
					goto nextelement;
				}
			}
			l.insert(l.end(), *it1);
		}
nextelement:;
	}
	v.assign(l.begin(), l.end());
}

void test_alignment(const std::string &a, const std::string &b) {
	int i=0, j=0;
	bool error = false;
	for (; !error && i<a.length() && j<b.length(); i++, j++) {
		while (a[i]=='-') ++i;
		while (b[j]=='-') ++j;
		if (a[i]!=b[j]) error = true;
	}
	while (a[i]=='-') ++i;
	while (b[j]=='-') ++j;
	if (i!=a.length() || j!=b.length()) error = true;
	if (error) {
		std::cerr << a << std::endl << b << std::endl;
		throw "Alignment wrong!";
	}
}

alignment make_alignment_jl(RNA &A, RNA &B, std::vector<match> &result, std::vector<double> &objectives) {
//	print_matches(result);
//	std::cout << std::endl;
	sort_matches(result);
//	print_matches(result);
//	std::cout << std::endl;
	int N1 = A.length;
	int N2 = B.length;
	int N = 0;
	for (std::vector<match>::iterator it=result.begin(); it!=result.end(); ++it) {
		N += (it->type==4 ? 2 :1);
	}
	std::string a(N, '-'), b(N, '-'), c(N, '.');
	int p = N;
	std::stack< std::pair<int, int> > stack;
	for (std::vector<match>::iterator it=result.begin(); it!=result.end(); ++it) {
		while (!stack.empty() && (
				it->type==1 && stack.top().first > it->j ||
				it->type==2 && stack.top().second > it->l ||
				it->type==3 && stack.top().first > it->j && stack.top().second > it->l ||
				it->type==4 && stack.top().first > it->j && stack.top().second > it->l
				)) {
//			std::cout << "pop (" << stack.top().first << "," << stack.top().second << ")" << std::endl;
			a[p-1] = A.sequence[stack.top().first-1];
			b[p-1] = B.sequence[stack.top().second-1];
			c[p-1] = '(';
			stack.pop();
			--p;
		}
		switch (it->type) {
			case 1:
				a[p-1] = A.sequence[it->j-1];
				break;
			case 2:
				b[p-1] = B.sequence[it->l-1];
				break;
			case 3:
				a[p-1] = A.sequence[it->j-1];
				b[p-1] = B.sequence[it->l-1];
				break;
			case 4:
				a[p-1] = A.sequence[it->j-1];
				b[p-1] = B.sequence[it->l-1];
				stack.push(std::pair<int,int>(it->i, it->k));
//				std::cout << "push (" << it->i << "," << it->k << ")" << std::endl;
				c[p-1] = ')';
				break;
		}
		--p;
	}
	while (!stack.empty()) {
//		std::cout << "pop (" << stack.top().first << "," << stack.top().second << ")" << std::endl;
		a[p-1] = A.sequence[stack.top().first-1];
		b[p-1] = B.sequence[stack.top().second-1];
		c[p-1] = '(';
		stack.pop();
		--p;
	}
	test_alignment(A.sequence, a);
	test_alignment(B.sequence, b);
	//std::cout << a << std::endl << b << std::endl << c << std::endl << std::endl;
	return alignment(objectives, a, b, c);

}
