#include <cstdio>
#include <vector>

#include <jni.h>

#include "alignment.hpp"
#include "rna.hpp"
#include "rna-pareto.hpp"
#include "rna-pareto-affine.hpp"
#include "rna-pareto-normal.hpp"

std::vector<alignment> find_pareto_alignments(const char* sequence_a, const char* sequence_b, bool affine, double gap_score, double gap_start) {
	std::vector<alignment> alignments;
	RNA A = RNA::fold_RNA(sequence_a);
	RNA B = RNA::fold_RNA(sequence_b);
	if (affine) {
		alignments = rna_pareto_affine::find_pareto_alignments(A, B, gap_score, gap_start);
	} else {
		alignments = rna_pareto_normal::find_pareto_alignments(A, B, gap_score);
	}
	return alignments;
}

/*
 * Class:     de_uniulm_inkombio_pareto_RNAPareto
 * Method:    getParetoAlignments
 * Signature: (Ljava/lang/String;Ljava/lang/String;ZDD)[Lde/uniulm/inkombio/pareto/Solution;
 */
extern "C" JNIEXPORT jobjectArray JNICALL Java_de_uniulm_inkombio_pareto_RNAPareto_getParetoAlignments(JNIEnv *env, jclass thisClass, jstring sequence_a, jstring sequence_b, jboolean affine, jdouble gap_score, jdouble gap_start) {
	const char *a = env->GetStringUTFChars(sequence_a, NULL);
	if (NULL == a) return NULL;
	const char *b = env->GetStringUTFChars(sequence_b, NULL);
	if (NULL == b) return NULL;
	bool baffine = (bool)(affine == JNI_TRUE);

	jclass solutionClass = env->FindClass("de/uniulm/inkombio/pareto/Solution");
	if (solutionClass==NULL) return NULL;
	jclass stringClass = env->FindClass("java/lang/String");
	if (stringClass==NULL) return NULL;

	std::vector<alignment> alignments = find_pareto_alignments(a, b, baffine, (double)gap_score, (double)gap_start);

	jobjectArray ret = (jobjectArray)env->NewObjectArray(alignments.size(), solutionClass, NULL);  
	jmethodID solutionConstructor = env->GetMethodID(solutionClass, "<init>", "([DLjava/lang/String;Ljava/lang/String;Ljava/lang/String;DD)V");
	if (solutionConstructor==NULL) return NULL;

	for (int i=0; i<alignments.size(); i++) {
		jdoubleArray objectives = env->NewDoubleArray(alignments[i].objectives.size());
		jdouble v[5];
		for (int j=0; j<alignments[i].objectives.size(); j++) v[j] = alignments[i].objectives[j];
		env->SetDoubleArrayRegion(objectives, 0, alignments[i].objectives.size(), v);
		jobject x = env->NewObject(solutionClass, solutionConstructor, 
				objectives,
				env->NewStringUTF(alignments[i].sequence_a.c_str()),
				env->NewStringUTF(alignments[i].sequence_b.c_str()), 
				env->NewStringUTF(alignments[i].structure.c_str()),
				alignment::energy_of_structure(alignments[i].sequence_a, alignments[i].structure),
				alignment::energy_of_structure(alignments[i].sequence_b, alignments[i].structure)
				);
		env->SetObjectArrayElement(ret, i, x);
	}
	env->ReleaseStringUTFChars(sequence_a, a);
	env->ReleaseStringUTFChars(sequence_b, b);
	return ret;
}

/*
 * Class:     de_uniulm_inkombio_pareto_RNAPareto
 * Method:    getVersion
 * Signature: ()Ljava/lang/String;
 */
extern "C" JNIEXPORT jstring JNICALL Java_de_uniulm_inkombio_pareto_RNAPareto_getVersion(JNIEnv *env, jclass thisClass) {
	char buffer[1024];
	int major = VERSION_MAJOR;
	int minor = VERSION_MINOR;
	int revision = VERSION_REVISION;
	sprintf(buffer, "%u.%d.%u", major, minor, revision);
	return env->NewStringUTF(buffer);
}

