#include <cstring>
#include <cmath>
#include <cfloat>

#include <algorithm>
#include <iostream>
#include <vector>
// Workaround to use <unordered_map> on MAC XCode
#ifndef IMPORT_TR1_UNORDERED_MAP
	#include <unordered_map>
#else
	#include <tr1/unordered_map>
	namespace std {
		using namespace tr1;
	}
#endif

#include "alignment.hpp"
#include "dpmatrix.hpp"
#include "rna.hpp"
#include "rna-pareto-normal.hpp"
#include "significant_base_pairs.hpp"
#include "sparsematrix.hpp"
#include "target.hpp"

using namespace std;

namespace rna_pareto_normal { // BEGIN namespace rna_pareto_normal

struct config {
	const double gap_score;
	const double p_min;

	const double epsilon;
	const double sigma;

	int maxsep;

	config() : maxsep(5), gap_score(-3.0), p_min(0.0005), epsilon(1e-8), sigma(0.05) {}

	config(double gap_score) : maxsep(5), gap_score(gap_score), p_min(0.0005), epsilon(1e-8), sigma(0.05) {}

	// sequence score: simple version from pmcomp
	double seq_score(const char a, const char b) const {
		return a==b ? sigma : 0.0;
	}
};

struct hash_pair_int_int {
	size_t operator()(const std::pair<int,int> &p) const {
		return std::hash<int>()(p.first) ^ std::hash<int>()(-p.second);
	}
};
typedef std::unordered_map<std::pair<int,int>, dpmatrix, hash_pair_int_int> memory_map;

// deletes all elements of v which do not belong to the pareto front
void pareto_clean(std::vector<target> &v) {
	std::vector<target> result;
	std::sort(v.begin(), v.end(), targetcomp);
	double max = -DBL_MAX;
	for (std::vector<target>::iterator it=v.begin(); it!=v.end(); ++it) {
		if (it->target_2 > max) {
			max = it->target_2;
			result.push_back(*it);
		}
	}
	std::swap(v, result);
}

void calculate_M_ik(RNA &A, RNA &B, significant_base_pairs &sig, dpmatrix &M_ik, sparsematrix<vector<target> > &D, int i, int k, config &config) {
	int N1 = A.length;
	int N2 = B.length;

	M_ik.clear();

	// initialization...
	for (int l=k+1; l<=N2+1 && l-k<=config.maxsep; l++) {
		M_ik[i][l].push_back(target((l-k) * config.gap_score, 0.0));
	}
	for (int j=i+1; j<=N1+1 && j-i<=config.maxsep; j++) {
		M_ik[j][k].push_back(target((j-i) * config.gap_score, 0.0));
	}
	M_ik[i][k].push_back(target(0.0, 0.0));

	if (i>0 && k>0 && !sig.bp[i][k]) return;

	// calculation...
	for (int j=i+1; j<=N1; j++) {
		for (int l=k+1; l<=N2; l++) {
			// comment out the following next/last statements to get the version
			// that restricts only pairs (not all subalignments)
			if ((l-k)-(j-i)>config.maxsep) break;
			if ((j-i)-(l-k)>config.maxsep) continue;
			if ((N1-j)-(N2-l)>config.maxsep) break;
			if ((N2-l)-(N1-j)>config.maxsep) continue;

			// calculate M[i,j,k,l];
			// case 1
			for (vector<target>::iterator it=M_ik[j-1][l-1].begin(); it!=M_ik[j-1][l-1].end(); ++it) {
				M_ik[j][l].push_back(target(it->target_1 + config.seq_score(A.sequence[j-1], B.sequence[l-1]), it->target_2, 1));
			}
			// case 2
			for (vector<target>::iterator it=M_ik[j-1][l].begin(); it!=M_ik[j-1][l].end(); ++it) {
				M_ik[j][l].push_back(target(it->target_1 + config.gap_score, it->target_2, 2));
			}
			// case 3
			for (vector<target>::iterator it=M_ik[j][l-1].begin(); it!=M_ik[j][l-1].end(); ++it) {
				M_ik[j][l].push_back(target(it->target_1 + config.gap_score, it->target_2, 3));
			}
			// case 4

			for (vector<int>::iterator it_j=upper_bound(sig.bplist_a[j].begin(), sig.bplist_a[j].end(), i); it_j!=sig.bplist_a[j].end(); ++it_j) { // *it_j : j'
			//for (vector<int>::iterator it_j=sig.bplist_a[j].begin(); it_j!=sig.bplist_a[j].end(); ++it_j) { // *it_j : j'
				//if (*it_j<=i) continue;
				if (*it_j>=j-3) break;
				for (vector<int>::iterator it_l=upper_bound(sig.bplist_b[l].begin(), sig.bplist_b[l].end(), k); it_l!=sig.bplist_b[l].end(); ++it_l) { // *it_l : l'
				//for (vector<int>::iterator it_l=sig.bplist_b[l].begin(); it_l!=sig.bplist_b[l].end(); ++it_l) { // *it_l : l'
					//if (*it_l<=k) continue;
					if (*it_l>=l-3) break;
					if (*it_l-*it_j>config.maxsep) break;
					if (*it_j-*it_l>config.maxsep) continue;
					//if (abs(*it_j-*it_l)>config.maxsep) continue;
					if ((*it_l-l)-(*it_j-j)>config.maxsep) break;
					if ((*it_j-j)-(*it_l-l)>config.maxsep) continue;
					//if (abs((*it_j-j)-(*it_l-l))>config.maxsep) continue;

//					cerr << "  case 4: j'=" << *it_j << " l'=" << *it_l << endl;
					for (vector<target>::iterator it1=M_ik[*it_j-1][*it_l-1].begin(); it1!=M_ik[*it_j-1][*it_l-1].end(); ++it1) {
						for (vector<target>::iterator it2=D(*it_j,j,*it_l,l).begin(); it2!=D(*it_j,j,*it_l,l).end(); ++it2) {
							M_ik[j][l].push_back(target(it1->target_1 + it2->target_1, it1->target_2 + it2->target_2, 4));
						}
					}
				}
			}
			pareto_clean(M_ik[j][l]);
		}
	}
}

void backtrack_M_ik(RNA &A, RNA &B, significant_base_pairs &sig, sparsematrix<vector<target> > &D, int i, int j, int k, int l, target t, vector<match> &result, memory_map &memory, config &config) {
	int N1 = A.length;
	int N2 = B.length;
//	cout << "backtrack [" << i << " " << j << " " << k << " " << l << "] type=" << t.type << endl;
	if (memory.count(pair<int,int>(i,k))==0) {
		memory[pair<int,int>(i,k)] = dpmatrix(N1+2, N2+2);
		dpmatrix &M_ik = memory[pair<int,int>(i,k)];
		calculate_M_ik(A, B, sig, M_ik, D, i, k, config);
	}

	while ((i<j) || (k<l)) {
		dpmatrix &M_ik = memory[pair<int,int>(i,k)];
		bool found = false;
		//cerr << "[" << i << " " << j << " " << k << " " << l << "] type=" << t.type << endl;
		switch (t.type) {
			case 0:
//				cerr << "zero encountered (" << i << " " << j << " " << k << " " << l << ")" << endl;
				if (i<j) {
					for (int h=j; h>i; h--) {
						result.push_back(match(1, h, l));
					}
				} else if (k<l) {
					for (int q=l; q>k; q--) {
						result.push_back(match(2, j, q));
					}
				} else {
					cerr << "ERROR: backtracking at zero, " << i << " " << j << " " << k << " " << l << endl;
				}
				return;
			case 1:
				if (M_ik[j-1][l-1].empty()) {
					cerr << "ERROR: vector " << i << " " << j-1 << " " << k << " " << l-1 << " empty" << endl;
				}
				for (vector<target>::iterator it=M_ik[j-1][l-1].begin(); it!=M_ik[j-1][l-1].end(); ++it) {
					if (abs(t.target_1 - (it->target_1 + config.seq_score(A.sequence[j-1], B.sequence[l-1])))<config.epsilon && abs(t.target_2 - it->target_2)<config.epsilon) {
						result.push_back(match(3, j, l));
//						cerr << "[" << j << " " << l << "]" << endl;
						--j;
						--l;
						t = *it;
						found = true;
						goto continuelabel;
					}
				}
				break;
			case 2:
				if (M_ik[j-1][l].empty()) {
					cerr << "ERROR: vector " << i << " " << j-1 << " " << k << " " << l << " empty" << endl;
				}
				for (vector<target>::iterator it=M_ik[j-1][l].begin(); it!=M_ik[j-1][l].end(); ++it) {
					if (abs(t.target_1 - (it->target_1 + config.gap_score))<config.epsilon && abs(t.target_2 - it->target_2)<config.epsilon) {
						result.push_back(match(1, j, l));
						--j;
						t = *it;
						found = true;
						goto continuelabel;
					}
				}
				break;
			case 3:
				if (M_ik[j][l-1].empty()) {
					cerr << "ERROR: vector " << i << " " << j << " " << k << " " << l-1 << " empty" << endl;
				}
				for (vector<target>::iterator it=M_ik[j][l-1].begin(); it!=M_ik[j][l-1].end(); ++it) {
					if (abs(t.target_1 - (it->target_1 + config.gap_score))<config.epsilon && abs(t.target_2 - it->target_2)<config.epsilon) {
						result.push_back(match(2, j, l));
						--l;
						t = *it;
						found = true;
						goto continuelabel;
					}
				}
				break;
			case 4:
				for (vector<int>::iterator it_j=sig.bplist_a[j].begin(); it_j!=sig.bplist_a[j].end(); ++it_j) { // *it_j : j'
					if (*it_j<=i) continue;
					if (*it_j>=j) break;
					for (vector<int>::iterator it_l=sig.bplist_b[l].begin(); it_l!=sig.bplist_b[l].end(); ++it_l) { // *it_l : l'
						if (*it_l<=k) continue;
						if (*it_l>=l) break;
						if (abs(*it_j-*it_l)>config.maxsep) continue;
						if (abs((*it_j-j)-(*it_l-l))>config.maxsep) continue;
//						cerr << "  case 4: j'=" << *it_j << " l'=" << *it_l << endl;
						for (vector<target>::iterator it1=M_ik[*it_j-1][*it_l-1].begin(); it1!=M_ik[*it_j-1][*it_l-1].end(); ++it1) {
							for (vector<target>::iterator it2=D(*it_j,j,*it_l,l).begin(); it2!=D(*it_j,j,*it_l,l).end(); ++it2) {
								if (abs(t.target_1 - (it1->target_1 + it2->target_1)) < config.epsilon && 
										abs(t.target_2 - (it1->target_2 + it2->target_2)) < config.epsilon) {
									result.push_back(match(4, *it_j, j, *it_l, l));
//									cerr << "[" << *it_j << " " << j << " " << *it_l << " " << l << "]" << endl;
//									cerr << "start searching: corresponding M["<<*it_j<<"]["<<j-1<<"]["<<*it_l<<"]["<<l-1<<"].size()="<<M[*it_j][j-1][*it_l][l-1].size() << ", D("<< *it_j<<","<<j<<","<<*it_l<<","<<l <<").size()=" << D(*it_j,j,*it_l,l).size() << endl;

									// perform backtracking
									backtrack_M_ik(A, B, sig, D, i, *it_j-1, k, *it_l-1, *it1, result, memory, config);

									// adjust variables, calc new matrix, resume work
									i = *it_j;
									k = *it_l;
									if (memory.count(pair<int,int>(i,k))==0) {
										memory[pair<int,int>(i,k)] = dpmatrix(N1+2,N2+2);
										dpmatrix &new_M_ik = memory[pair<int,int>(i,k)];
										calculate_M_ik(A, B, sig, new_M_ik, D, i, k, config);
									}
									dpmatrix &new_M_ik = memory[pair<int,int>(i,k)];

									bool found2 = false;
									for (vector<target>::iterator it3=new_M_ik[j-1][l-1].begin(); it3!=new_M_ik[j-1][l-1].end(); ++it3) {
										if (abs(it2->target_1 - it3->target_1) < config.epsilon && 
												abs(it2->target_2 - (it3->target_2+A.psi[i][j]+B.psi[k][l])) < config.epsilon) {
											t = *it3;
											found2 = true;
											break;
										}
									}	
									if (!found2) {
										cerr << "ERROR: no corresponding M elemend to D(" << *it_j << ", " << j << ", " << *it_l << ", " << l << ") found, backtracking failed!!!" << endl;
									}
									j = j-1;
									l = l-1;
//									cout << "resume [" << i << " " << j << " " << k << " " << l << "] type=" << t.type << endl;
									found = true;
									goto continuelabel;
								}
							}
						}
					}
				}
				break;
			default:
				cerr << "error, type was " << t.type << endl;
				break;
		}
continuelabel:
		if (!found) {
			cerr << "Error: backtrack failed at " << i << " " << j << " " << k << " " << l << ", type was " << t.type << endl;
//			exit(1);
			return;
		}
	}
}

vector<alignment> find_pareto_alignments(RNA &A, RNA &B, double gap_score) {
	int N1 = A.length;
	int N2 = B.length;
	// cerr << "N1 = " << N1 << ", N2 = " << N2 << endl;
	config config(gap_score);
	if (config.maxsep < abs(N1-N2)){
		cerr << "increasing maxsep to difference in sequence length" << endl;
		config.maxsep = abs(N1-N2);
	}

	significant_base_pairs sig(A, B, config.p_min);
	sparsematrix<vector<target> > D_ik;
	dpmatrix M_ik(N1+2, N2+2);

	// calculation
	for (int i=N1; i>=0; i--) {
		for (int k=N2; k>=0; k--) {

			// comment out the following next/last statements to get the version
			// that restricts only pairs (not all subalignments)
			if (i-k>config.maxsep) break;
			if (k-i>config.maxsep) continue;
			if (i>0 && k>0 && !sig.bp[i][k]) continue;

			calculate_M_ik(A, B, sig, M_ik, D_ik, i, k, config);
			for (int j=i+1; j<=N1; j++) {
				for (int l=k+1; l<=N2; l++) {
					if (i>0 && k>0 && A.bppm[i][j]>=config.p_min && B.bppm[k][l]>=config.p_min) {
						vector<target> &v_ik = D_ik(i,j,k,l);
						for (vector<target>::iterator it=M_ik[j-1][l-1].begin(); it!=M_ik[j-1][l-1].end(); ++it) {
							v_ik.push_back(target(it->target_1, it->target_2 + A.psi[i][j] + B.psi[k][l], 4));
						}
					}
				}
			}
		}
	}
//	cerr << "M[0][" << N1 << "][0]["<<N2<<"].size()="<<M[0][N1][0][N2].size()<<endl;
	vector<alignment> alignments;
	memory_map memory;
	memory[pair<int,int>(0,0)] = M_ik;
	for (vector<target>::iterator it=M_ik[N1][N2].begin(); it!=M_ik[N1][N2].end(); ++it) {
		//printf("%lf %lf %lf\n", it->target_1, it->target_2, it->target_1+it->target_2);
		vector<match> matches;
		backtrack_M_ik(A, B, sig, D_ik, 0, N1, 0, N2, *it, matches, memory, config);
		double tmp[3] = {it->target_1, it->target_2, it->target_1+it->target_2};
		vector<double> values(tmp, tmp+3);
		alignments.push_back(make_alignment_jl(A, B, matches, values));
	}
	return alignments;
}

} // END namespace rna_pareto_normal
